import * as React from 'react';
import { Route, Routes, Navigate, useLocation } from 'react-router-dom';
import { IApp, IPlatformApp } from '@cloudbase/cals';
import { default as PageRender, IAppProps } from './index';
import { getComponentSchema, IComponentInfo } from '../utils/adapter/generate';
import { NotFound } from './status';
import { generatePageRelativePath } from '../utils/common';

export type TCALS = IApp | IPlatformApp;

interface IPackageProps
  extends Pick<
    IAppProps,
    'generateComponentsMap' | 'generateActionsMap' | 'onRender' | 'initDataContext' | 'renderLoading'
  > {
  cals: TCALS;
  componentsInfoMap: Record<string, IComponentInfo>;
  lowcodeModuleMap: {
    [pageId: string]: any;
  };
  historyOption?: { _indexPage?: boolean };
  onPageMounted?: IAppProps['onMounted'];
  pageOptions?: IAppProps['options'];
  onNotFound?: Function;
}

export function PackageRender({
  cals,
  componentsInfoMap,
  lowcodeModuleMap,
  historyOption,
  generateComponentsMap,
  generateActionsMap,
  onPageMounted,
  pageOptions,
  onRender,
  initDataContext,
  renderLoading,
  onNotFound,
}: IPackageProps) {
  const { items } = cals;
  const home = React.useMemo(() => {
    return ((items || []) as any[]).find((item) => item.id === cals.main);
  }, [items, cals.main]);
  return (
    <Routes>
      {home && (
        <Route
          path="/"
          element={<Navigate to={`.${generatePageRelativePath(home, historyOption?._indexPage)}`} replace={true} />}
        />
      )}
      {items.map((page) => {
        const uuid = cals.extra.rootPath ? `${cals.extra.rootPath}/${page.id}` : page.id;
        return (
          <Route
            key={page.id}
            path={generatePageRelativePath(page, historyOption?._indexPage)}
            element={
              <PageRender
                page={{ ...page, uuid }}
                generateComponentsMap={generateComponentsMap}
                generateActionsMap={generateActionsMap}
                generateLowcodeModule={() => {
                  console.log('lowcodeModuleMap', lowcodeModuleMap);
                  return lowcodeModuleMap[uuid] || {};
                }}
                generateContext={(ctx, page, options) => {
                  const { widgets, componentSchema, dataBinds } = getComponentSchema(
                    ctx,
                    page,
                    componentsInfoMap,
                    undefined,
                    {
                      isComposite: false,
                      looseError: options.looseError,
                      processCssUnit: options.processCssUnit,
                    },
                  );
                  return {
                    widgetContext: widgets,
                    componentSchema,
                    dataBinds,
                  };
                }}
                onMounted={onPageMounted}
                onRender={onRender}
                initDataContext={initDataContext}
                options={pageOptions}
                renderLoading={renderLoading}
              />
            }
          ></Route>
        );
      })}
      <Route path="/1" element={<MockPage />}></Route>

      <Route path="*" element={<NotFound onNotFound={onNotFound} />}></Route>
    </Routes>
  );
}

function MockPage() {
  const location = useLocation();
  return (
    <>
      <p>subpackage path</p>
      <p>{location.pathname}</p>
    </>
  );
}
